part of gauges;

/// Create radial gauge widget to displays numerical values on a circular scale.
/// It has a rich set of features
/// such as axes, ranges, pointers, and annotations that are fully
/// customizable and extendable.
/// Use it to create speedometers, temperature monitors, dashboards,
/// meter gauges, multiaxis clocks, watches, activity gauges, compasses,
/// and more.
///
/// The radial gauge widget allows to customize its appearance
/// using [SfGaugeThemeData] available from [SfGaugeTheme] widget or
/// the [SfTheme] with the help of [SfThemeData].
///
/// ```dart
/// Widget build(BuildContext context) {
///    return Container(
///        child: SfRadialGauge());
///}
/// ```
/// See also:
///
/// * [SfTheme] and [SfThemeData], for customizing the visual appearance
/// of the radial gauge.
/// * [SfGaugeTheme] and [SfGaugeThemeData], for customizing the visual
/// appearance of the radial gauge.
class SfRadialGauge extends StatefulWidget {
  /// Creates a radial gauge with default or required axis.
  ///
  /// To enable the loading animation set [enableLoadingAnimation] is true.
  // ignore: prefer_const_constructors_in_immutables
  SfRadialGauge(
      {Key key,
      List<RadialAxis> axes,
      this.enableLoadingAnimation = false,
      this.animationDuration = 2000,
      this.title})
      : axes = axes ?? <RadialAxis>[RadialAxis()],
        super(key: key);

  /// Add a list of radial axis to the gauge and customize each axis by
  /// adding it to the [axes] collection.
  ///
  /// Also refer [RadialAxis]
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfRadialGauge(
  ///          axes:<RadialAxis>[RadialAxis(
  ///            )]
  ///        ));
  ///}
  /// ```
  final List<RadialAxis> axes;

  /// Add the title to the top of the radial gauge.
  ///
  /// The [title] is a type of [GaugeTitle].
  ///
  /// To specify title description to [GaugeTitle.text] property and
  /// title style to [GaugeTitle.textStyle] property.
  ///
  /// Defaults to null.
  ///
  /// Also refer [GaugeTitle].
  ///
  ///
  /// ```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfRadialGauge(
  ///            title: GaugeTitle(
  ///                    text: 'Gauge Title',
  ///                    textStyle: TextStyle(
  ///                                 color: Colors.red,
  ///                                 fontSize: 12,
  ///                                 fontStyle: FontStyle.normal,
  ///                                 fontWeight: FontWeight.w400,
  ///                                 fontFamily: 'Roboto'
  ///                               ))
  ///        ));
  ///}
  /// ```
  final GaugeTitle title;

  /// Specifies the load time animation for axis elements, range and
  /// pointers with [animationDuration].
  ///
  /// Defaults to false
  ///
  /// ```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfRadialGauge(enableLoadingAnimation: true
  ///        ));
  ///}
  /// ```
  final bool enableLoadingAnimation;

  /// Specifies the load time animation duration.
  ///
  /// Duration is defined in milliseconds.
  ///
  /// Defaults to 2000
  /// ```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfRadialGauge(animationDuration: 4000
  ///        ));
  ///}
  /// ```
  final double animationDuration;

  @override
  State<StatefulWidget> createState() {
    return _SfRadialGaugeState();
  }
}

/// Represents the radial gauge state
class _SfRadialGaugeState extends State<SfRadialGauge>
    with SingleTickerProviderStateMixin {
  /// Represents the gauge theme
  SfGaugeThemeData _gaugeTheme;

  /// Hold the radial gauge animation details
  _AnimationDetails _animationDetails;

  @override
  void initState() {
    _animationDetails = _AnimationDetails();
    _animationDetails.pointerRepaintNotifier = ValueNotifier<int>(0);
    _animationDetails.rangeRepaintNotifier = ValueNotifier<int>(0);
    _animationDetails.axisRepaintNotifier = ValueNotifier<int>(0);
    _animationDetails.animationController = AnimationController(vsync: this)
      ..addListener(_repaintGaugeElements);
    _animationDetails.needsToAnimateAxes =
        widget.enableLoadingAnimation && widget.animationDuration > 0;
    _animationDetails.needsToAnimateRanges =
        widget.enableLoadingAnimation && widget.animationDuration > 0;
    _animationDetails.needsToAnimatePointers =
        widget.enableLoadingAnimation && widget.animationDuration > 0;
    _animationDetails.needsToAnimateAnnotation =
        widget.enableLoadingAnimation && widget.animationDuration > 0;
    if (widget.axes != null && widget.axes.isNotEmpty) {
      for (num i = 0; i < widget.axes.length; i++) {
        final RadialAxis axis = widget.axes[i];
        if (axis.pointers != null && axis.pointers.isNotEmpty) {
          for (num j = 0; j < axis.pointers.length; j++) {
            final GaugePointer pointer = axis.pointers[j];
            pointer._needsAnimate = true;
            pointer._animationStartValue =
                axis.isInversed && !(pointer is RangePointer) ? 1 : 0;
          }
        }
      }
    }
    super.initState();
  }

  @override
  void didUpdateWidget(SfRadialGauge oldWidget) {
    if (widget.axes != null && widget.axes.isNotEmpty) {
      _animationDetails.needsToAnimateAnnotation = false;
      _animationDetails.needsToAnimatePointers = false;
      _animationDetails.needsToAnimateRanges = false;
      _animationDetails.needsToAnimateAxes = false;
      _needsRepaintGauge(oldWidget, widget);
    }
    super.didUpdateWidget(oldWidget);
  }

  @override
  void didChangeDependencies() {
    _gaugeTheme = SfGaugeTheme.of(context);
    super.didChangeDependencies();
  }

  /// Method to repaint the gauge elements
  void _repaintGaugeElements() {
    if (_animationDetails.needsToAnimateAxes) {
      _animationDetails.axisRepaintNotifier.value++;
    }
    if (_animationDetails.needsToAnimateRanges) {
      _animationDetails.rangeRepaintNotifier.value++;
    }
    _animationDetails.pointerRepaintNotifier.value++;
  }

  /// Method to check whether the gauge needs to be repainted
  void _needsRepaintGauge(SfRadialGauge oldGauge, SfRadialGauge newGauge) {
    if (oldGauge.axes != null &&
        newGauge.axes != null &&
        oldGauge.axes.length == newGauge.axes.length) {
      for (num i = 0; i < newGauge.axes.length; i++) {
        _needsRepaintAxis(oldGauge.axes[i], newGauge.axes[i]);
      }
    } else {
      for (num i = 0; i < newGauge.axes.length; i++) {
        newGauge.axes[i]._needsRepaintAxis = true;
      }
    }
  }

  /// Method to check whether the axis needs to be repainted
  void _needsRepaintAxis(RadialAxis oldAxis, RadialAxis newAxis) {
    if (oldAxis.backgroundImage == newAxis.backgroundImage &&
        oldAxis._backgroundImageInfo?.image != null) {
      newAxis._backgroundImageInfo = oldAxis._backgroundImageInfo;
    }
    if (newAxis != oldAxis) {
      newAxis._needsRepaintAxis = true;
    } else {
      newAxis._needsRepaintAxis = false;
    }
    if (newAxis.pointers != null) {
      _needsRepaintPointer(oldAxis, newAxis);
    }
    if (newAxis.ranges != null) {
      _needsRepaintRange(oldAxis, newAxis);
    }
  }

  /// Checks whether to repaint the ranges
  void _needsRepaintRange(RadialAxis oldAxis, RadialAxis newAxis) {
    if (newAxis.ranges != null) {
      if (oldAxis.ranges != null) {
        if (newAxis.ranges.length == oldAxis.ranges.length) {
          for (num i = 0; i < newAxis.ranges.length; i++) {
            final GaugeRange newRange = newAxis.ranges[i];
            final GaugeRange oldRange = oldAxis.ranges[i];
            if (newRange != oldRange) {
              newRange._needsRepaintRange = true;
            } else {
              if (newAxis._needsRepaintAxis) {
                newRange._needsRepaintRange = true;
              } else {
                newRange._needsRepaintRange = false;
              }
            }
          }
        } else {
          for (num i = 0; i < newAxis.ranges.length; i++) {
            newAxis.ranges[i]._needsRepaintRange = true;
          }
        }
      } else {
        for (num i = 0; i < newAxis.ranges.length; i++) {
          newAxis.ranges[i]._needsRepaintRange = true;
        }
      }
    }
  }

  /// Checks whether to repaint the axis pointers
  void _needsRepaintPointer(RadialAxis oldAxis, RadialAxis newAxis) {
    if (newAxis.pointers != null) {
      if (oldAxis.pointers != null) {
        if (newAxis.pointers.length == oldAxis.pointers.length) {
          _needsAnimatePointers(oldAxis, newAxis);
          _needsResetPointerValue(oldAxis, newAxis);
          for (num i = 0; i < newAxis.pointers.length; i++) {
            final GaugePointer newPointer = newAxis.pointers[i];
            final GaugePointer oldPointer = oldAxis.pointers[i];
            if (newPointer != oldPointer) {
              newPointer._needsRepaintPointer = true;
            } else {
              if (newAxis._needsRepaintAxis) {
                newPointer._needsRepaintPointer = true;
              } else {
                newPointer._needsRepaintPointer = false;
              }
            }
          }
        } else {
          for (num i = 0; i < newAxis.pointers.length; i++) {
            newAxis.pointers[i]._needsRepaintPointer = true;
          }
        }
      }
    }
  }

  /// Checks whether to animate the pointers
  void _needsAnimatePointers(RadialAxis oldAxis, RadialAxis newAxis) {
    for (num i = 0; i < newAxis.pointers.length; i++) {
      if (oldAxis.pointers[i].value != newAxis.pointers[i].value) {
        setState(() {
          // Sets the previous animation end value as current animation start
          // value of pointer
          newAxis.pointers[i]._needsAnimate = true;
          newAxis.pointers[i]._animationStartValue =
              oldAxis.pointers[i]._animationEndValue;
        });
      } else if (oldAxis.pointers[i].animationType !=
          newAxis.pointers[i].animationType) {
        newAxis.pointers[i]._needsAnimate = true;
      } else {
        setState(() {
          newAxis.pointers[i]._needsAnimate = false;
        });
      }
    }
  }

  /// Check to reset the pointer current value
  void _needsResetPointerValue(RadialAxis oldAxis, RadialAxis newAxis) {
    for (num i = 0; i < newAxis.pointers.length; i++) {
      if (oldAxis.pointers[i].enableDragging ==
          newAxis.pointers[i].enableDragging) {
        if (oldAxis.pointers[i].value == newAxis.pointers[i].value &&
            newAxis.pointers[i].value == newAxis.pointers[i]._currentValue &&
            oldAxis.pointers[i]._currentValue !=
                newAxis.pointers[i]._currentValue) {
          newAxis.pointers[i]._currentValue = oldAxis.pointers[i]._currentValue;
        }
        newAxis.pointers[i]._isDragStarted = oldAxis.pointers[i]._isDragStarted;
      }
    }
  }

  /// Methods to add the title of circular gauge
  Widget _addGaugeTitle() {
    if (widget.title != null && widget.title.text.isNotEmpty) {
      final Widget titleWidget = Container(
          child: Container(
              padding: const EdgeInsets.fromLTRB(0, 5, 0, 0),
              decoration: BoxDecoration(
                  color: widget.title.backgroundColor ??
                      _gaugeTheme.titleBackgroundColor,
                  border: Border.all(
                      color: widget.title.borderColor ??
                          _gaugeTheme.titleBorderColor,
                      width: widget.title.borderWidth)),
              child: Text(
                widget.title.text,
                style: widget.title.textStyle,
                textAlign: TextAlign.center,
                overflow: TextOverflow.clip,
              ),
              alignment: (widget.title.alignment == GaugeAlignment.near)
                  ? Alignment.topLeft
                  : (widget.title.alignment == GaugeAlignment.far)
                      ? Alignment.topRight
                      : (widget.title.alignment == GaugeAlignment.center)
                          ? Alignment.topCenter
                          : Alignment.topCenter));

      return titleWidget;
    } else {
      return Container();
    }
  }

  /// Method to add the elements of gauge
  Widget _addGaugeElements() {
    return Expanded(child: LayoutBuilder(
      builder: (BuildContext context, BoxConstraints constraints) {
        return _AxisContainer(widget, _gaugeTheme, _animationDetails);
      },
    ));
  }

  @override
  Widget build(BuildContext context) {
    SyncfusionLicense.validateLicense(context);
    return LimitedBox(
        maxHeight: 350,
        maxWidth: 350,
        child:
            Column(children: <Widget>[_addGaugeTitle(), _addGaugeElements()]));
  }

  @override
  void dispose() {
    if (_animationDetails.animationController != null) {
      _animationDetails.animationController
          .removeListener(_repaintGaugeElements);
      _animationDetails.animationController.dispose();
    }

    if (widget.axes != null && widget.axes.isNotEmpty) {
      for (int i = 0; i < widget.axes.length; i++) {
        final RadialAxis axis = widget.axes[i];
        if (axis._imageStream != null) {
          axis._imageStream.removeListener(axis._listener);
        }
      }
    }
    super.dispose();
  }
}

/// Holds the animation and repainter details.
class _AnimationDetails {
  /// Holds the pointer repaint notifier
  ValueNotifier<int> pointerRepaintNotifier;

  /// Holds the range repaint notifier
  ValueNotifier<int> rangeRepaintNotifier;

  /// Holds the axis repaint notifier
  ValueNotifier<int> axisRepaintNotifier;

  /// Holds the animation controller
  AnimationController animationController;

  /// Specifies whether to animate axes
  bool needsToAnimateAxes;

  /// Specifies whether to animate ranges
  bool needsToAnimateRanges;

  /// Specifies whether to animate pointers
  bool needsToAnimatePointers;

  /// Specifies whether to animate annotation
  bool needsToAnimateAnnotation;
}
